#!/usr/bin/env python3
import gc
import weakref

from rbloom import Bloom
from hashlib import sha256
from pickle import dumps
import os


def test_bloom(bloom: Bloom):
    assert not bloom
    assert bloom.approx_items == 0.0

    bloom.add('foo')
    assert bloom
    assert bloom.approx_items > 0.0

    bloom.add('bar')

    assert 'foo' in bloom
    assert 'bar' in bloom
    assert 'baz' not in bloom

    bloom.update(['baz', 'qux'])
    assert 'baz' in bloom
    assert 'qux' in bloom

    other = bloom.copy()
    assert other == bloom
    assert other is not bloom

    other.clear()
    assert not other
    assert other.approx_items == 0.0

    other.update(['foo', 'bar', 'baz', 'qux'])
    assert other == bloom

    other.update(str(i).encode()*500 for i in range(100000))
    for i in range(100000):
        assert str(i).encode()*500 in other
    assert bloom != other
    assert bloom & other == bloom
    assert bloom | other == other

    bloom &= other
    assert bloom < other

    orig = bloom.copy()
    bloom |= other
    assert bloom == other
    assert bloom > orig
    assert bloom >= orig
    assert bloom.issuperset(other)
    assert orig <= bloom
    assert orig.issubset(bloom)
    assert bloom >= bloom
    assert bloom.issuperset(bloom)
    assert bloom <= bloom
    assert bloom.issubset(bloom)

    bloom = orig.copy()
    bloom.update(other)
    assert bloom == other
    assert bloom > orig

    bloom = orig.copy()
    assert other == bloom.union(other)
    assert bloom == bloom.intersection(other)

    bloom.intersection_update(other)
    assert bloom == orig

    # TEST PERSISTENCE
    if not bloom.hash_func is hash:
        # find a filename that doesn't exist
        i = 0
        while os.path.exists(f'UNIT_TEST_{i}.bloom'):
            i += 1
        filename = f'test{i}.bloom'

        try:
            # save and load
            bloom.save(filename)
            bloom2 = Bloom.load(filename, bloom.hash_func)
            assert bloom == bloom2
        finally:
            # remove the file
            os.remove(filename)

        # TEST bytes PERSISTENCE
        bloom_bytes = bloom.save_bytes()
        assert type(bloom_bytes) == bytes
        bloom3 = Bloom.load_bytes(bloom_bytes, bloom.hash_func)
        assert bloom == bloom3


def sha_based(obj):
    h = sha256(dumps(obj)).digest()
    return int.from_bytes(h[:16], "big") - 2**127


def circular_ref():
    def loop_hash_func(x):
        return sha_based(x)
    weak_ref = weakref.ref(loop_hash_func)
    bloom = Bloom(1000, 0.1, hash_func=loop_hash_func)
    assert gc.get_referents(bloom) == [loop_hash_func]
    loop_hash_func.bloom = bloom
    del bloom
    del loop_hash_func
    gc.collect()
    assert weak_ref() is None

def operations_with_self():
    bloom = Bloom(1000, 0.1)
    bloom.add('foo')
    assert 'foo' in bloom
    bloom |= bloom
    bloom &= bloom
    bloom.update(bloom)
    bloom.update(bloom, bloom)
    bloom.update(bloom, ['bob'], bloom)
    assert 'foo' in bloom
    assert 'bob' in bloom

    bloom.intersection_update(bloom)
    bloom.intersection_update(bloom, bloom)
    bloom.intersection_update(bloom, ['foo'], bloom)
    assert 'foo' in bloom
    assert 'bob' not in bloom
    assert bloom == bloom
    assert bloom <= bloom
    assert bloom >= bloom
    assert not (bloom > bloom)
    assert not (bloom < bloom)
    assert bloom.issubset(bloom)
    assert bloom.issuperset(bloom)
    assert bloom.union(bloom) == bloom
    assert bloom.union(bloom, bloom) == bloom
    assert bloom.intersection(bloom) == bloom
    assert bloom.intersection(bloom, bloom) == bloom


def api_suite():
    assert repr(Bloom(27_000, 0.0317)) == "<Bloom size_in_bits=193960 approx_items=0.0>"
    assert Bloom(1140, 0.999).hash_func == hash
    assert Bloom(102, 0.01, hash_func=hash).hash_func is hash
    assert Bloom(103100, 0.51, hash_func=sha_based).hash_func is sha_based

    test_bloom(Bloom(13242, 0.0000001))
    test_bloom(Bloom(9874124, 0.01, hash_func=sha_based))
    test_bloom(Bloom(2837, 0.5, hash_func=hash))

    circular_ref()
    operations_with_self()

    print('All API tests passed')


if __name__ == '__main__':
    api_suite()



import struct
import timeit

import pybloomfilter
import bloom_filter2
import flor
import pybloom
import rbloom


NUM_ITEMS = 10_000_000


def run(ty):
    bf = ty(NUM_ITEMS, 0.01)

    for i in range(NUM_ITEMS):
        bf.add(i + 0.5)  # floats because ints are hashed as themselves

    for i in range(NUM_ITEMS):
        if i + 0.5 not in bf:
            raise ValueError("Should be no false negatives")


def run_bytes(ty):
    bf = ty(NUM_ITEMS, 0.01)

    for i in range(NUM_ITEMS):
        bf.add(struct.pack("d", i + 0.5))

    for i in range(NUM_ITEMS):
        if struct.pack("d", i + 0.5) not in bf:
            raise ValueError("Should be no false negatives")


types = {
    "rbloom": rbloom.Bloom,
    "pybloomfiltermmap3": pybloomfilter.BloomFilter,
    "pybloom3": pybloom.BloomFilter,
    "flor": flor.BloomFilter,
    "bloomfilter2": bloom_filter2.BloomFilter,
}


def main():
    for name, ty in types.items():
        print(f"Running {name}")
        try:
            results = timeit.repeat(lambda: run(ty), number=1, repeat=5)
            extras = ""
        except Exception as e:
            results = timeit.repeat(lambda: run_bytes(ty), number=1, repeat=5)
            extras = f" (via bytes because {e})"
        avg = sum(results) / len(results)
        print(f"  {avg:6.2f}s{extras}")


if __name__ == "__main__":
    main()



import os
from typing import Any, Callable, Iterable, Union, final


@final
class Bloom:

    # expected_items:  max number of items to be added to the filter
    # false_positive_rate:  max false positive rate of the filter
    # hash_func:  optional argument, see section "Cryptographic security"
    def __init__(self, expected_items: int, false_positive_rate: float,
                 hash_func=__builtins__.hash) -> None: ...

    # number of buckets in the filter
    @property
    def size_in_bits(self) -> int: ...

    # retrieve the hash_func given to __init__
    @property
    def hash_func(self) -> Callable[[Any], int]: ...

    # estimated number of items in the filter
    @property
    def approx_items(self) -> float: ...

    # load from file, see section "Persistence"
    @classmethod
    def load(cls, filepath: Union[str, bytes, os.PathLike], hash_func: Callable[[Any], int]) -> Bloom: ...

    # load from bytes(), see section "Persistence"
    @classmethod
    def load_bytes(cls, data: bytes, hash_func: Callable[[Any], int]) -> Bloom: ...

    # save to file, see section "Persistence"
    def save(self, filepath: Union[str, bytes, os.PathLike]) -> None: ...

    # save to a bytes(), see section "Persistence"
    def save_bytes(self) -> bytes: ...

    #####################################################################
    #                    ALL SUBSEQUENT METHODS ARE                     #
    #              EQUIVALENT TO THE CORRESPONDING METHODS              #
    #                     OF THE BUILT-IN SET TYPE                      #
    #####################################################################

    def add(self, obj: Any, /) -> None: ...

    def __contains__(self, obj: Any) -> bool: ...

    def __bool__(self) -> bool: ...                   # False if empty

    def __repr__(self) -> str: ...                    # basic info

    def __or__(self, other: Bloom) -> Bloom: ...      # self | other

    def __ior__(self, other: Bloom) -> None: ...      # self |= other

    def __and__(self, other: Bloom) -> Bloom: ...     # self & other

    def __iand__(self, other: Bloom) -> None: ...     # self &= other

    # extension of __or__
    def union(self, *others: Union[Iterable, Bloom]) -> Bloom: ...

    # extension of __ior__
    def update(self, *others: Union[Iterable, Bloom]) -> None: ...

    # extension of __and__
    def intersection(self, *others: Union[Iterable, Bloom]) -> Bloom: ...

    # extension of __iand__
    def intersection_update(self, *others: Union[Iterable, Bloom]) -> None: ...

    # these implement <, >, <=, >=, ==, !=
    def __lt__(self, other: Bloom) -> bool: ...
    def __gt__(self, other: Bloom) -> bool: ...
    def __le__(self, other: Bloom) -> bool: ...
    def __ge__(self, other: Bloom) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    def issubset(self, other: Bloom, /) -> bool: ...      # self <= other

    def issuperset(self, other: Bloom, /) -> bool: ...    # self >= other

    def clear(self) -> None: ...                          # remove all items

    def copy(self) -> Bloom: ...                          # duplicate self



import timeit

import time

NUMBER = 1000000


def format_time(time_ns: float) -> str:
    return f"{time_ns / 1000:.04} us"


def main():
    res = timeit.timeit(
        setup=f"from rbloom import Bloom; b = Bloom({NUMBER}, 0.01)",
        stmt="b.add(object())",
        timer=time.perf_counter_ns,
        number=NUMBER,
    )
    print("Time to insert an element:")
    print(format_time(res / NUMBER))

    results = timeit.repeat(
        setup=f"from rbloom import Bloom; b = Bloom({NUMBER}, 0.01); objects = [object() for _ in range({NUMBER})]",
        stmt="b.update(objects)",
        timer=time.perf_counter_ns,
        number=1,
        repeat=20,
    )
    res = min(results)
    print("Time to insert each element in a batch:")
    print(format_time(res / NUMBER))

    results = timeit.repeat(
        setup=f"from rbloom import Bloom; b = Bloom({NUMBER}, 0.01); objects = (object() for _ in range({NUMBER}))",
        stmt="b.update(objects)",
        timer=time.perf_counter_ns,
        number=1,
        repeat=20,
    )
    res = min(results)
    print("Time to insert each element in a batch via an iterable:")
    print(format_time(res / NUMBER))

    res = timeit.timeit(
        setup=f"from rbloom import Bloom; b = Bloom({NUMBER}, 0.01); stored_obj = object(); b.add(stored_obj); b.update(object() for _ in range({NUMBER}))",
        stmt="stored_obj in b",
        timer=time.perf_counter_ns,
        number=NUMBER,
    )
    print("Time to check if an object is present:")
    print(format_time(res / NUMBER))


if __name__ == "__main__":
    main()
